# -*- test-case-name: nevow.test.test_appserver -*-

import cgi
from copy import copy
from urllib import unquote
from types import StringType

from twisted.web import server
from twisted.web import resource
from twisted.web.woven import guard
from twisted.web import util as webutil

from twisted.protocols import http

from twisted.python import log
from twisted.python import failure
from twisted.internet import defer
from twisted.application import service

from nevow import compy
from nevow import context
from nevow import inevow
from nevow import events
from nevow import url
from nevow import util
from nevow import flat


class PrettyPrinter(compy.Adapter):
    def getResourceForRequest(self, request):
        name = request.postpath[-1]
        if name == '$$HTMLSource':
            request.postpath.pop()
            collector = []
            oldwrite = request.write
            oldfinish = request.finish
            def interceptWrite(written):
                collector.append(written)
            def interceptFinish():
                source = ''.join(collector)
                from twisted.web.microdom import parseString
                tree = parseString(source)
                textized = tree.toprettyxml()
                oldwrite('<html><body><pre>')
                oldwrite(util.escapeToXML(textized))
                oldwrite('</pre></body></html>')
                oldwrite(textized)
                oldfinish()
            request.write = interceptWrite
            request.finish = interceptFinish


class DefaultExceptionHandler:
    __implements__ = inevow.ICanHandleException

    def renderHTTP_exception(self, request, reason):
        request.write("<html><head><title>Exception</title></head><body>")
        from nevow import failure
        result = failure.formatFailure(reason)
        request.write(''.join(flat.flatten(flat.serialize(result, context.WovenContext()))))
        request.write("</body></html>")


exceptionHandler = DefaultExceptionHandler()


class NevowRequest(compy.Componentized, server.Request):
    __implements__ = compy.IComponentized,
    """A Request subclass which does additional
    processing if a form was POSTed. When a form is POSTed,
    we create a cgi.FieldStorage instance using the data posted,
    and set it as the request.fields attribute. This way, we can
    get at information about filenames and mime-types of
    files that were posted.
    """

    def __init__(self, *args, **kw):
        server.Request.__init__(self, *args, **kw)
        compy.Componentized.__init__(self)

    def processingFailed(self, reason):
        try:
            if self.contexts:
                ctx = self.contexts[-1]
            else:
                ctx = self.context

            handler = ctx.locate(inevow.ICanHandleException)
            return handler.renderHTTP_exception(self, reason)
        except:
            return self._reallyFailed(reason)

    def _reallyFailed(self, reason):
        self.setResponseCode(http.INTERNAL_SERVER_ERROR)
        log.err(reason)
        self.write("<html><head><title>Traceback from Nevow</title></head><body>")
        self.write(webutil.formatFailure(reason))
        self.write("</body></html>")
        return None

    def process(self):
        # extra request parsing
        if self.method == 'POST':
            t = self.content.tell()
            self.content.seek(0)
            self.fields = cgi.FieldStorage(self.content, self.received_headers,
                                           environ={'REQUEST_METHOD': 'POST'})
            self.content.seek(t)

        # get site from channel
        self.site = self.channel.site

        # set various default headers
        self.setHeader('server', server.version)
        self.setHeader('date', server.http.datetimeToString())
        self.setHeader('content-type', "text/html")

        # context setup
        #session = self.getSession()
        #session.context = context.WovenContext()

        self.context = context.RequestContext(tag=self)
        #self.context.remember(exceptionHandler, inevow.ICanHandleException)

        # Resource Identification
        self.prepath = []
        self.postpath = map(unquote, self.path[1:].split('/'))
        self.deferred = defer.Deferred()
        return self.site.getResourceFor(
            self
        ).addCallback(
            lambda resource: inevow.IResource(resource).renderHTTP(self)
        ).addErrback(
            self.processingFailed
        ).addBoth(
            self._cbFinishRender
        )

    def finish(self):
        self.deferred.callback("")

    def _cbFinishRender(self, html):
        if compy.implements(html, inevow.IResource) or compy.implements(html, resource.IResource):
            # allow returning a Resource from renderHTTP.
            
            ## FIXME!
            ## This is nasty code duplication from appserver.handleSegment
            
            newres = html
            ## Keep track of where we found this guy and keep going
            self.site.url.objectMountedAt(newres, self.prepath)
            
            ## Create a context object to represent this new resource
            ctx = context.WovenContext(tag=newres, parent=self.contexts[-1])
            self.contexts.append(ctx)
            newres.context = ctx
            self.resources.append(newres)

            ## FIXME! And this is nasty code duplication from .process!
            return defer.maybeDeferred(
                inevow.IResource(newres).renderHTTP, self
            ).addErrback(
                self.processingFailed
            ).addBoth(
                self._cbFinishRender
            )
        
        elif isinstance(html, str):
            self.write(html)
        else:
            print "html is not a string: ", str(html)
        server.Request.finish(self)
        return html

    session = None

    def getSession(self, sessionInterface=None):
        if self.session is not None:
            self.session.touch()
            if sessionInterface:
                return sessionInterface(self.session)
            return self.session
        ## temporary until things settle down with the new sessions
        return server.Request.getSession(self, sessionInterface)

    def URLPath(self):
        return url.URLPath.fromRequest(self)


def sessionFactory(ctx):
    """Given a RequestContext instance with a Request as .tag, return a session
    """
    return ctx.tag.getSession()

requestFactory = lambda ctx: ctx.tag


class OldResourceAdapter(object):
    __implements__ = inevow.IResource,

    def __init__(self, original):
        self.original = original

    def __repr__(self):
        return "<%s @ 0x%x adapting %r>" % (self.__class__.__name__, id(self), self.original)

    def locateChild(self, request, segments):
        if self.original.isLeaf:
            return self, ()
        name = segments[0]
        request.prepath.append(request.postpath.pop(0))
        res = self.original.getChildWithDefault(name, request)
        request.postpath.insert(0, request.prepath.pop())
        if isinstance(res, defer.Deferred):
            return res.addCallback(lambda res: (res, segments[1:]))
        return res, segments[1:]

    def _handle_NOT_DONE_YET(self, data, request):
        if data == server.NOT_DONE_YET:
            return request.deferred
        else:
            return data
        
    def renderHTTP(self, request):
        result = defer.maybeDeferred(self.original.render, request).addCallback(
            self._handle_NOT_DONE_YET, request)
        return result

    def willHandle_notFound(self, request):
        if hasattr(self.original, 'willHandle_notFound'):
            return self.original.willHandle_notFound(request)
        return False

    def renderHTTP_notFound(self, request):
        return self.original.renderHTTP_notFound(request)

    def _getContext(self):
        return getattr(self.original, 'context', None)

    def _setContext(self, new):
        setattr(self.original, 'context', new)

    context = property(
        _getContext, 
        _setContext)


class NevowSite(server.Site):
    """A Site subclass which does additional pre-processing before
    delegating the request rendering to a Resource instance.
    By default, this Site is able to deal with Guard login/logout
    requests, Freeform form posts, and pretty-printing the HTML
    source of any page.
    """
    requestFactory = NevowRequest
    generatorFactories = [
        PrettyPrinter,
    ]
    def __init__(self, *args, **kw):
        server.Site.__init__(self,
            #logPath="web.log",
            *args, **kw)
        self.notification = events.EventNotification()
        self.url = url.URLGenerator()
        self.resourceGenerators = [
            x(self) for x in self.generatorFactories
        ]

    def getResourceFor(self, request):
        """Retrieve a resource from this site for a particular request, first
        checking all resource generators.
        """
        request.site = self
        request.sitepath = copy(request.prepath)
        res = None
        for gener in self.resourceGenerators:
            res = gener.getResourceForRequest(request)
            if res is not None:
                break
        if res is not None:
            if isinstance(res, defer.Deferred):
                return res
            return defer.succeed(res)
        return self.resourceAtPathOfRequest(request)

    def resourceAtPathOfRequest(self, request):
        """Retrieve a resource from this site for a particular request,
        ignoring all generators and looking only at the path from the default
        resource.  (This is used primarily from within resource generators to
        locate the default resource which they are wrapping.)
        """
        request.resources = []
        request.contexts = []
        path = tuple(request.postpath)
        res = inevow.IResource(self.resource)
        request.contexts.append(context.PageContext(tag=res, parent=request.context))
        return defer.maybeDeferred(res.locateChild, request, path).addCallback(
            self.handleSegment, request, path, res
        ).addCallback(
            self.finishRemember
        )

    def handleSegment(self, newresult, request, path, res):
        newres, newpath = newresult
        newres = inevow.IResource(newres, persist=True)
        if newres is res:
            assert not newpath is path, "URL traversal cycle detected when attempting to locateChild %r from resource %r." % (path, res)
            assert  len(newpath) < len(path), "Infinite loop impending..."

        ## We found a Resource... update the request.prepath and postpath
        for x in xrange(len(path) - len(newpath)):
            request.prepath.append(request.postpath.pop(0))

        ## Keep track of where we found this guy and keep going
        self.url.objectMountedAt(newres, request.prepath)

        ## Create a context object to represent this new resource
        ctx = context.PageContext(tag=newres, parent=request.contexts[-1])
        request.contexts.append(ctx)
        newres.context = ctx
        request.resources.append(newres)

        res = newres
        path = newpath

        if not path:
            return newres

        return defer.maybeDeferred(res.locateChild, request, path).addCallback(
            self.handleSegment, request, path, res
        )

    def finishRemember(self, res):
        if hasattr(res, 'remember'):
            #res.remember(self.url, url.IURLGenerator)
            #res.remember(self.notification, events.IEventMaster)
            pass
        return res

