# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from __future__ import generators

import warnings

from nevow import compy
from nevow.inevow import IData, IRequest
from nevow.stan import specialMatches, specials, Tag, Unset
from nevow.tags import invisible
from nevow.util import qual

class _Marker(object):
    pass


def megaGetInterfaces(adapter):
    adrs = [qual(x) for x in compy.getInterfaces(adapter)]
    ## temporarily turn this off till we need it
    if False: #hasattr(adapter, '_adapterCache'):
        adrs.extend(adapter._adapterCache.keys())
    return adrs


dataqual = qual(IData)


class NodeNotFound(KeyError):
    def __str__(self):
        return "The %s named %r wasn't found in the template." % tuple(self.args[:2])

class TooManyNodes(Exception):
    def __str__(self):
        return "More than one %r with the name %r was found." % tuple(self.args[:2])

class WovenContext(object):
    key = None
    _remembrances = None
    tag = None
    _slotData = None
    isAttrib = False
    def __init__(self, parent=None, tag=None, precompile=False, remembrances=None, key='', isAttrib=False):
        self.tag = tag
        sd = getattr(tag, 'slotData', None)
        if sd is not None:
            self._slotData = sd
        self.parent = parent
        if key is not None and key is not Unset:
            if self.parent is not None and self.parent.key:
                self.key = '.'.join((self.parent.key, key))
            else:
                self.key = key
            #print "KEY", `self.key`
        else:
            ## Bubble the value down to the bottom so it's always immediately accessible
            if self.parent is not None:
                self.key = self.parent.key
        self._remembrances = remembrances
        self.precompile = precompile
        self.isAttrib = isAttrib

    def __repr__(self):
        rstr = ''
        if self._remembrances:
            rstr = ', remembrances=%r' % self._remembrances
        attribstr=''
        if self.isAttrib:
            attribstr="isAttrib"
        return "%s(tag=%r%s%s)" % (self.__class__.__name__, self.tag, rstr,attribstr)

    def remember(self, adapter, interface=None):
        """Remember an object that implements some interfaces.
        Later, calls to .locate which are passed an interface implemented
        by this object will return this object.
        
        If the 'interface' argument is supplied, this object will only
        be remembered for this interface, and not any of
        the other interfaces it implements.
        """
        if interface is None:
            interfaceList = megaGetInterfaces(adapter)
            if not interfaceList:
                interfaceList = [dataqual]
        else:
            interfaceList = [qual(interface)]
        if self._remembrances is None:
            self._remembrances = {}
        for interface in interfaceList:
            self._remembrances[interface] = adapter
        return self

    def with(self, tag):
        """Usually for nevow machinery use only, unless you know what you are doing.

        Create a new context stack entry for the specified tag.
        
        tag: the tag the new context stack entry will be associated with.
        """

        new = WovenContext(self, tag, self.precompile, key=tag.key, isAttrib=self.isAttrib)
        if tag.remember is not Unset:
            new.remember(tag.remember)
        return new

    def locate(self, interface, depth=1):
        """Locate an object which implements a given interface.
        Objects will be searched through the context stack top
        down.
        """
        key = qual(interface)
        if depth < 0:
            full = []
            while True:
                try:
                    full.append(self.locate(interface, len(full)+1))
                except KeyError:
                    break
            #print "full", full, depth
            if full:
                return full[depth]
            return None
        if self._remembrances is not None and self._remembrances.has_key(key):
            depth -= 1
            if not depth:
                return self._remembrances[key]
        if self.parent is None:
            raise KeyError, "Interface %s was not remembered." % key
        return self.parent.locate(interface, depth)

    def chain(self, context):
        """For nevow machinery use only.

        Go to the top of this context's context chain, and make
        the given context the parent, thus continuing the chain
        into the given context's chain.
        """
        top = self
        while top.parent is not None:
            if top.parent.tag is None:
                ## If top.parent.tag is None, that means this context (top)
                ## is just a marker. We want to insert the current context
                ## (context) as the parent of this context (top) to chain properly.
                break
            top = top.parent
            if top is context: # this context is already in the chain; don't create a cycle
                return
        top.parent = context

    def patterns(self, pattern, default=None):
        warnings.warn("use patternGenerator instead", stacklevel=2)
        return self.patternGenerator(pattern, default)
    
    def patternGenerator(self, pattern, default=None):
        """Returns a psudeo-Tag which will generate clones of matching
        pattern tags forever, looping around to the beginning when running
        out of unique matches.
        
        If no matches are found, and default is None, raise an exception,
        otherwise, generate clones of default forever.

        You can use the normal stan syntax on the return value.
        
        Useful to find repeating pattern elements. Example rendering function:
        def simpleSequence(context, data):
          pattern = context.patternCloner('item')
          return [pattern(data=element) for element in data]
        """
        patterner = self._locatePatterns(self.tag, pattern, default)
        return PatternTag(patterner)

    def allPatterns(self, pattern):
        """Return a list of all matching pattern tags, cloned.
        
        Useful if you just want to insert them in the output in one
        place.
        
        E.g. the sequence renderer's header and footer are found with this.
        """
        return [tag.clone(deep=False) for tag in 
                specialMatches(self.tag, 'pattern', pattern)]

    def onePattern(self, pattern):
        """Return a single matching pattern, cloned.
        If there is more than one matching pattern or no matching patterns,
        raise an exception.

        Useful in the case where you want to locate one and only one
        sub-tag and do something with it.
        """
        return self._locateOne(pattern,
                               lambda pattern: specialMatches(self.tag, 'pattern', pattern),
                               'pattern').clone(deep=False)

    def fillSlots(self, name, stan):
        """Set 'stan' as the stan tree to replace all slots with name 'name'.
        """
        if self._slotData is None:
            self._slotData = {}
        self._slotData[name] = stan

    def locateSlotData(self, name):
        """For use by nevow machinery only, or for some fancy cases.

        Find previously remembered slot filler data.
        For use by flatstan.SlotRenderer"""
        if self._slotData:
            data = self._slotData.get(name, Unset)
            if data is not Unset:
                return data
        if self.parent is None:
            raise KeyError, "Slot named '%s' was not filled." % name
        return self.parent.locateSlotData(name)
    
    def _locatePatterns(self, tag, pattern, default):
        keeplooking = True
        gen = specialMatches(tag, 'pattern', pattern)
        produced = []
        while keeplooking:
            keeplooking = False
            for x in gen or produced:
                if gen:
                    produced.append(x)
                keeplooking = True
                cloned = x.clone(deep=False)
                cloned.pattern = Unset
                yield cloned
            gen=None
        if default is None:
            raise NodeNotFound, "Pattern %s was not found." % pattern
        if hasattr(default, 'clone'):
            while True:  yield default.clone(deep=False)
        else:
            while True:  yield default
                
    def _locateOne(self, name, locator, descr):
        found = False
        for node in locator(name):
            if found:
                raise TooManyNodes(descr, name)
            found = node
        if not found:
            raise NodeNotFound(descr, name)
        return found

    def clone(self, deep=True, cloneTags=True):
        ## don't clone the tags of parent contexts. I *hope* code won't be
        ## trying to modify parent tags so this should not be necessary.
        ## However, *do* clone the parent contexts themselves.
        ## This is necessary for chain(), as it mutates top-context.parent.
        
        if self.parent:
            parent=self.parent.clone(cloneTags=False)
        else:
            parent=None
        if cloneTags:
            tag = self.tag.clone(deep=deep)
        else:
            tag = self.tag
        if self._remembrances is not None:
            remembrances=self._remembrances.copy()
        else:
            remembrances=None
        return WovenContext(
            parent = parent,
            tag = tag,
            remembrances=remembrances,
            isAttrib=self.isAttrib
        )

    def arg(self, get, default=None):
        """Placeholder until I can find Jerub's implementation of this

        Return a single named argument from the request arguments
        """
        req = self.locate(IRequest)
        return req.args.get(get, [default])[0]


class FactoryContext(WovenContext):
    """A context which allows adapters to be registered against it so that an object 
    can be lazily created and returned at render time. When ctx.locate is called
    with an interface for which an adapter is registered, that adapter will be used
    and the result returned.
    """
    cache = None
    def locate(self, interface, depth=1):
        if self.cache is None: self.cache = {}
        adapter = self.cache.get(interface, None)
        if adapter is not None:
            return adapter
        adapter = interface(self, default=None)
        if adapter is not None:
            self.cache[interface] = adapter
            return adapter
        return WovenContext.locate(self, interface, depth)


class RequestContext(FactoryContext):
    """A RequestContext has adapters for the following interfaces:
    
    ISession
    IFormDefaults
    IFormErrors
    IHand
    IStatusMessage
    """
    pass


class PageContext(FactoryContext):
    """A PageContext has adapters for the following interfaces:

    IRenderer
    IRendererFactory
    IData
    """
    pass


class PatternTag(object):
    '''A pseudotag created by WovenContext.patternGenerator() which loops
    through a sequence of matching patterns.'''

    def __init__(self, patterner):
        self.pat = patterner.next()
        self.patterner = patterner

    def __call__(self, **kw):
        if self.pat:
            p, self.pat = self.pat, None
            return p(**kw)
        return self.patterner.next()(**kw)

    def __getitem__(self, item):
        if self.pat:
            p, self.pat = self.pat, None
            return p[item]
        return self.patterner.next()[item]
