import os.path

from nevow import context
from nevow import inevow
from nevow import tags


class DocFactory:
    """Base class for document factories. I am responsible for loading a document
    template for the renderer to use.
    """

    __implements__ = inevow.IDocFactory,
    
    pattern = None
    precompiledDoc = None

    def __init__(self, pattern=None):
        if pattern:
            self.pattern = pattern

    def precompile(self):
        ctx = context.WovenContext(precompile=True)
        from nevow import flat
        doc = flat.precompile(self.getDoc(), ctx)
        if self.pattern:
            ctx = ctx.with(tags.invisible[doc])
            doc = [ctx.onePattern(self.pattern)]
        return doc

    def load(self):
        """Load and return the document for the renderer"""
        if self.precompiledDoc is None:
            self.precompiledDoc = self.precompile()
        return self.precompiledDoc


# Caches for the document factories
_stanCache = {}
_htmlstrCache = {}
_htmlfileCache = {}
_xmlstrCache = {}
_xmlfileCache = {}


class stan(DocFactory):
    """A stan tags document factory"""

    def __init__(self, doc, **kw):
        self.doc = doc
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        return self.doc


class htmlstr(DocFactory):
    """A document factory for HTML contained in a string"""

    beExtremelyLenient = False

    def __init__(self, template, beExtremelyLenient=None, **kw):
        self.template = template
        if beExtremelyLenient:
            self.beExtremelyLenient = beExtremelyLenient
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from twisted.web import microdom
        return microdom.parseString(
            self.template,
            beExtremelyLenient=self.beExtremelyLenient
            )

    def precompile(self):
        doc = _htmlstrCache.get((self.template, self.pattern), None)
        if doc:
            return doc
        doc = DocFactory.precompile(self)
        _htmlstrCache[(self.template, self.pattern)] = doc
        return doc


class htmlfile(DocFactory):
    """A document factory for an HTML disk template"""

    template = None
    templateDir = ''
    beExtremelyLenient = False
    _mtime = None

    def __init__(self, template, templateDir=None, beExtremelyLenient=None, **kw):
        self.template = template
        if templateDir is not None:
            self.templateDir = templateDir
        if beExtremelyLenient is not None:
            self.beExtremelyLenient = beExtremelyLenient
        self._filename = os.path.join(self.templateDir, self.template)
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from twisted.web import microdom
        return microdom.parse(
            self._filename,
            beExtremelyLenient=self.beExtremelyLenient
            )

    def precompile(self):
        mtime, doc = _htmlfileCache.get((self._filename, self.pattern), (None,None))
        self._mtime = os.path.getmtime(self._filename)
        if mtime == self._mtime:
            return doc
        doc = DocFactory.precompile(self)
        _htmlfileCache[(self._filename, self.pattern)] = self._mtime, doc
        return doc
        
    def load(self):
        if self._mtime != os.path.getmtime(self._filename):
            self.precompiledDoc = None
        return DocFactory.load(self)


class xmlstr(DocFactory):

    doc = None

    def __init__(self, template, **kw):
        self.template = template
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from nevow.flat import flatsax
        return flatsax.parseString(self.template)

    def precompile(self):
        doc = _xmlstrCache.get((self.template, self.pattern), None)
        if doc:
            return doc
        doc = DocFactory.precompile(self)
        _xmlstrCache[(self.template, self.pattern)] = doc
        return doc
        

class xmlfile(DocFactory):

    template = None
    templateDir = ''

    _mtime = None

    def __init__(self, template, templateDir=None, **kw):
        self.template = template
        if templateDir:
            self.templateDir = templateDir
        self._filename = os.path.join(self.templateDir, self.template)
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from nevow.flat import flatsax
        return flatsax.parse(file(self._filename))
    
    def precompile(self):
        mtime, doc = _xmlfileCache.get((self._filename, self.pattern), (None,None))
        self._mtime = os.path.getmtime(self._filename)
        if mtime == self._mtime:
            return doc
        doc = DocFactory.precompile(self)
        _xmlfileCache[(self._filename, self.pattern)] = self._mtime, doc
        return doc
        
    def load(self):
        if self._mtime != os.path.getmtime(self._filename):
            self.precompiledDoc = None
        return DocFactory.load(self)
