from cStringIO import StringIO
import os.path
import urllib
import warnings
import traceback

from twisted.python import log
from twisted.python.context import get as getCtx

from nevow.context import WovenContext, NodeNotFound, PageContext, RequestContext
from nevow import compy
from nevow import inevow
from nevow import liveevil
from nevow import tags
from nevow import flat
from nevow.util import Deferred, maybeDeferred

import formless
from formless import iformless

from time import time as now



# For backward compatability, provide lazy docFactories
def deprecationWrapper(name):
    def _(*args, **kwargs):
        warnings.warn(
            "rend.%s has been deprecated. Use loaders.%s instead." % (name, name),
            DeprecationWarning, 2)
        from nevow import loaders
        wrapped = getattr(loaders, name)
        return wrapped(*args, **kwargs)
    return _

htmlfile = deprecationWrapper('htmlfile')
htmlstr = deprecationWrapper('htmlstr')
stan = deprecationWrapper('stan')
xmlfile = deprecationWrapper('xmlfile')
xmlstr = deprecationWrapper('xmlstr')


class RenderFactory(object):
    __implements__ = inevow.IRendererFactory,

    def renderer(self, context, name):
        """Return a renderer with the given name.
        """

        # The named renderer can be parameterised, i.e. 'renderIt one,two,three'
        args = []
        if name.find(' ') != -1:
            name, args = name.split(None, 1)
            args = [arg.strip() for arg in args.split(',')]
        
        callable = getattr(self, 'render_%s' % name, None)
        if callable is None:
            callable = lambda context, data: context.tag[
                "The renderer named '%s' was not found in %r." % (name, self)]

        if args:
            return callable(*args)

        return callable

    render_sequence = lambda self, context, data: sequence(context, data)
    render_mapping = lambda self, context, data: mapping(context, data)
    render_string = lambda self, context, data: context.tag.clear()[str(data)]
    render_xml = lambda self, context, data: context.tag.clear()[tags.xml(data)]
    render_remove = lambda self, context, data: ''
    render_mergeSlot = lambda self, slot: mergeSlot(slot)


class DataFactory(object):
    __implements__ = inevow.IContainer,

    def child(self, context, name):
        callable = getattr(self, 'data_%s' % name, None)
        if callable is None:
            container = inevow.IContainer(self.original, None)
            if container is None:
                callable = lambda context, data: context.tag["The data named '%s' was not found in %r." % (name, self)]
            else:
                callable = container.child
        return callable


# TODO:
#
# Remove dependence on t.w.resource.IResource interface
# Add childFactory?
# Allow doc factories to be rendered, for instance:
#     def render_foo(c, d): return span(data=person, render=htmlstr(...))
# More unit tests:
#   * URL traversal


class LiveEvilChildMixin:
    """Mixin that provides the LiveEvil child resources."""
    child_nevow_liveOutput = liveevil.liveOutput
    child_nevow_liveInput = liveevil.liveInput


class FreeformChildMixin:
    """Mixin that handles locateChild for freeform segments."""
    def locateChild(self, request, segments):
        ## The method or property name we are going to validate against/affect
        bindingName = None

        name = segments[0]
        if name.startswith('freeform_post!'):
            configurableName, bindingName = name.split('!')[1:3]
        elif name.startswith('freeform-action-post!'):
            configurableName, request.args['freeform-actee'] = name.split('!')[1:3]
            bindingName = request.args['freeform-action'][0]
             
        if bindingName:
            ctx = self.context
            if ctx is None:
                ctx = request.context
                self.context = ctx
            ctx.remember(self, inevow.IResource)
            ctx.remember(request, inevow.IRequest)
            cf = iformless.IConfigurableFactory(self)
            c = cf.locateConfigurable(ctx, configurableName)
            return self.webFormPost(request, self, c, ctx, bindingName, request.args)
        return NotFound


class ConfigurableFactory:
    """Locates configurables by looking for methods that start with
    configurable_ and end with the name of the configurable. The method
    should take a single arg (other than self) - the current context.
    """
    __implements__ = iformless.IConfigurableFactory

    def locateConfigurable(self, context, name):
        return iformless.IConfigurable(getattr(self, 'configurable_%s'%name)(context))

    def configurable_(self, context):
        if compy.implements(self, iformless.IConfigurable):
            return self
        if self.original:
            if compy.implements(self.original, formless.TypedInterface):
                return self.original
        return self


_CARRYOVER = {}


def defaultsFactory(ctx):
    co = _CARRYOVER.get(
        ctx.tag.args.get('_nevow_carryover_', [None])[0], None)
    from formless import webform
    defaults = webform.FormDefaults()
    if co is not None:
        e = iformless.IFormErrors(co, {})
        for k, v in e.items():
            defaults.getAllDefaults(k).update(v.partialForm)
    return defaults


def errorsFactory(ctx):
    co = _CARRYOVER.get(
        ctx.tag.args.get('_nevow_carryover_', [None])[0], None)
    from formless import webform
    errs = webform.FormErrors()
    if co is not None:
        e = iformless.IFormErrors(co, {})
        for k, v in e.items():
            errs.updateErrors(k, v.errors)
            errs.setError(k, v.formErrorMessage)
    return errs


def handFactory(ctx):
    co = _CARRYOVER.get(
        ctx.tag.args.get('_nevow_carryover_', [None])[0], None)
    return inevow.IHand(co, None)


def statusFactory(ctx):
    co = _CARRYOVER.get(
        ctx.tag.args.get('_nevow_carryover_', [None])[0], None)
    return inevow.IStatusMessage(co, None)


def originalFactory(ctx):
    return ctx.tag



class PageGettable(compy.Adapter):
    def get(self, context):
        return self.original.original


def deferflatten(stan, ctx, writer):
    finished = Deferred()

    iterable = flat.iterflatten(
        stan, ctx, writer, lambda D: isinstance(D, Deferred))

    def drive():
        try:
            deferred, returner = iterable.next()
            def cb(result):
                returner(result)
                drive()
                return result
            def eb(failure):
                finished.errback(failure)
                return failure
            cfac = getCtx('CursorFactory')
            if cfac:
                deferred.addCallback(cfac.store.transback, cb).addErrback(cfac.store.transback, eb)
            else:
                deferred.addCallback(cb).addErrback(eb)
        except StopIteration:
            finished.callback('')

    drive()
    return finished

def deferFlattenFactory(p, stan, ctx, writer, finisher):
    return deferflatten(stan, ctx, writer).addCallback(finisher)


class Page(FreeformChildMixin, LiveEvilChildMixin, RenderFactory, DataFactory, ConfigurableFactory):
    """A page is the main Nevow resource and renders a document loaded
    via the document factory (docFactory). A page can also be used as a
    component that is embedded in another page.
    """

    __implements__ = inevow.IRenderer, RenderFactory.__implements__,inevow.IResource, DataFactory.__implements__, ConfigurableFactory.__implements__

    original = None
    docFactory = None
    buffered = False

    beforeRender = None
    afterRender = None
    addSlash = None

    flattenFactory = deferFlattenFactory
    
    _context = None

    def __init__(self, original=None, docFactory=None):
        if original:
            self.original = original
        if docFactory:
            self.docFactory = docFactory
        self.toremember = []
        self.children = {}
        self.remember(self, iformless.IConfigurableFactory)

    def putChild(self, name, child):
        self.children[name] = child

    def getParentContext(self, request):
        contexts = getattr(request, 'contexts', [])
        if contexts:
            return contexts[-1]
        return None

    ##
    # IResource methods
    ##

    def locateChild(self, request, segments):
        """Locate a child page of this one. Request is a
        nevow.appserver.NevowRequest, and segments is a tuple of each element
        in the URI. An tuple (page, segments) should be returned, with page as
        in instance of nevow.rend.Page and segments a tuple representing the
        remaining segments of the URI. If the child is not found, return
        rend.NotFound instead.

        This looks for children in three places, in this order:

        - in a dictionary, self.children
        - a member of self named child_<childname>. This can be either an
          attribute or a method. If an attribute, it should be an object which
          can be adapted to IResource. If a method, it should take the request
          and return an instance which can be adapted to IResource.
        - by calling self.getDynamicChild(name, request). Name is a string.
          This should return an instance that can be adapted to IResource.
        """

        r = self.children.get(segments[0], None)
        if r:
            return r, segments[1:]

        w = getattr(self, 'child_%s'%segments[0], None)
        if w:
            from twisted.web import resource
            if compy.implements(w, inevow.IResource) or compy.implements(w, resource.IResource):
                return w, segments[1:]
            return w(request), segments[1:]

        r = self.getDynamicChild(segments[0], request)
        if r:
            return r, segments[1:]

        return FreeformChildMixin.locateChild(self, request, segments)

    def getDynamicChild(self, segment, request):
        """Used by locateChild to return children which are generated
        dynamically. Note that higher level interfaces use only locateChild,
        and only nevow.rend.Page.locateChild uses this.

        segment is a string represnting one element of the URI. Request is a
        nevow.appserver.NevowRequest.

        The default implementation of this always returns None; it is intended
        to be overridden."""
        return None

    def renderHTTP(self, request):
        if self.beforeRender is not None:
            self.beforeRender(request)
        if self.addSlash and request.prepath[-1] != '':
            request.redirect(request.URLPath().child(''))
            return ''
            
        log.msg(http_render=None, uri=request.uri)

        self.remember(request, inevow.IRequest)
        self.rememberStuff(self)

        def finishRequest():
            carryover = request.args.get('_nevow_carryover_', [None])[0]
            if carryover is not None and _CARRYOVER.has_key(carryover):
                del _CARRYOVER[carryover]
            self.context = None
            if self.afterRender is not None:
                self.afterRender(request)

        if self.buffered:
            io = StringIO()
            writer = io.write
            def finisher(result):
                request.write(io.getvalue())
                finishRequest()
                return result
        else:
            writer = request.write
            def finisher(result):
                finishRequest()
                return result

        if self.context is None:
            ## We were created outside of the normal locateChild lookup
            ## loop, and thus don't have any context to which we are attached.
            ## create one.
            self.context = PageContext(tag=self, parent=RequestContext(tag=request))
        assert self.context is not None

        doc = self.docFactory.load()
        ctx =  WovenContext(self.context, tags.invisible[doc])
        
        return self.flattenFactory(doc, ctx, writer, finisher)

    def rememberStuff(self, ctx):
        ctx.remember(self, inevow.IResource)
        ctx.remember(self, inevow.IRenderer)
        ctx.remember(self, inevow.IRendererFactory)

        if self.original is None:
            ctx.remember(self, inevow.IData)
        else:
            ctx.remember(self.original, inevow.IData)
            ctx.remember(self.original)

    def renderString(self):

        io = StringIO()
        writer = io.write

        def finisher(result):
            return io.getvalue()
        
        if self.context is None:
            self.context = PageContext(tag=self)

        doc = self.docFactory.load()
        ctx =  WovenContext(self.context, tags.invisible[doc])
        self.rememberStuff(ctx)

        return self.flattenFactory(doc, ctx, writer, finisher)

    def child_(self, request):
        return self

    def webFormPost(self, request, res, configurable, ctx, bindingName, args):
        def redirectAfterPost(aspects):
            redirectAfterPost = request.getComponent(iformless.IRedirectAfterPost, None)
            if redirectAfterPost is None:
                ref = request.getHeader('referer') or ''
                from nevow import url
                refpath = url.URL.fromString(ref)
                magicCookie = str(now())
                refpath = refpath.replace('_nevow_carryover_', magicCookie)
                _CARRYOVER[magicCookie] = C = compy.Componentized(aspects)
                request.redirect(str(refpath))
            else:
                ## Use the redirectAfterPost url
                request.redirect(str(redirectAfterPost))
            from nevow import static
            return static.Data('You posted a form to %s' % bindingName, 'text/plain'), ()
        return maybeDeferred(
            configurable.postForm, ctx, bindingName, args
        ).addCallback(
            self.onPostSuccess, request, ctx, bindingName, redirectAfterPost
        ).addErrback(
            self.onPostFailure, request, ctx, bindingName, redirectAfterPost
        )

    def onPostSuccess(self, result, request, ctx, bindingName, redirectAfterPost):

        if result is None:
            message = "Operation '%s' was completed successfully." % \
                      formless.nameToLabel(bindingName)
        else:
            message = result
            
        return redirectAfterPost({inevow.IHand: result, inevow.IStatusMessage: message})

    def onPostFailure(self, reason, request, ctx, bindingName, redirectAfterPost):
        reason.trap(formless.ValidateError)
        return redirectAfterPost({iformless.IFormErrors: {bindingName: reason.value}})

    def remember(self, obj, inter=None):
        """Remember the given object for the given interfaces (or all interfaces
        obj.__implements__) in the top-level context for this page.

        Note that the objects remembered using this API are only remembered
        for the duration of one render.
        """
        if self.context is None:
            self.toremember.append((obj, inter))
        else:
            ## We have a context. Remember this immediately.
            self.context.remember(obj, inter)

    def _setContext(self, new):
        self._context = new
        if new is not None:
            for rememberit in self.toremember:
                new.remember(*rememberit)

    context = property(
        lambda self: self._context,
        _setContext)

    def rend(self, context, data):
        context.remember(self, inevow.IRenderer)
        context.remember(self, inevow.IRendererFactory)
        context.remember(self, inevow.IData)

        # This tidbit is to enable us to include Page objects inside
        # stan expressions and render_* methods and the like. But
        # because of the way objects can get intertwined, we shouldn't
        # leave the pattern changed.
        old = self.docFactory.pattern
        self.docFactory.pattern = 'content'
        self.docFactory.precompiledDoc = None
        try:
            doc = self.docFactory.load()
            self.docFactory.pattern = old
            self.docFactory.precompiledDoc = None
        except NodeNotFound:
            self.docFactory.pattern = old
            self.docFactory.precompiledDoc = None
            doc = self.docFactory.load()
        return doc


def sequence(context, data):
    """Renders each item in the sequence using patterns found in the
    children of the element.

    Sequence recognises the following patterns:

    header: Rendered at the start, before the first item. If multiple
        header patterns are provided they are rendered together in the
        order they were defined.
            
    footer: Just like the header only renderer at the end, after the
        last item.
    
    item: Rendered once for each item in the sequence. If multiple
        item patterns are provided then the pattern is cycled in the
        order defined.
        
    divider: Rendered once between each item in the sequence. Multiple
        divider patterns are cycled.

    Example:
    
    <table nevow:render="sequence" nevow:data="peopleSeq">
      <tr nevow:pattern="header">
        <th>name</th>
        <th>email</th>
      </tr>
      <tr nevow:pattern="item" class="odd">
        <td>name goes here</td>
        <td>email goes here</td>
      </tr>
      <tr nevow:pattern="item" class="even">
        <td>name goes here</td>
        <td>email goes here</td>
      </tr>
    </table>
            
    """
    headers = context.allPatterns('header')
    pattern = context.patternGenerator('item')
    divider = context.patternGenerator('divider', default=tags.invisible)
    content = [(pattern(data=element), divider(data=element)) for element in data]
    if not content:
        content = context.allPatterns('empty')
    else:
        ## No divider after the last thing.
        content[-1] = content[-1][0]
    footers = context.allPatterns('footer')
    
    return context.tag.clear()[ headers, content, footers ]


def mapping(context, data):
    """Fills any slots in the element's children with data from a
    dictionary. The dict keys are used as the slot names, the dict
    values are used as filling.

    Example:

    <tr nevow:render="mapping" nevow:data="personDict">
      <td><nevow:slot name="name"/></td>
      <td><nevow:slot name="email"/></td>
    </tr>
    """
    for k, v in data.items():
        context.fillSlots(k, v)
    return context.tag


def mergeSlot(slot):
    def _(context, data):
        mergeTag = context.locateSlotData(slot)
        context.tag(**mergeTag.attributes)
        if mergeTag.children:
            context.tag.clear()[mergeTag.children]
        return context.tag
    return _


class FourOhFour(Page):
    """A simple 404 (not found) page.
    """
    isLeaf = True
    notFound = "<html><head><title>Page Not Found</title><head><body>Sorry, but I couldn't find the object you requested.</body></html>"

    def renderHTTP(self, request):
        from twisted.protocols import http
        request.setResponseCode(http.NOT_FOUND)
        if request.contexts:
            ctx = request.contexts[-1]
        else:
            ctx = request.context
        try:
            return ctx.locate(inevow.ICanHandleNotFound).renderHTTP_notFound(request)
        except KeyError, e:
            return self.notFound
        except Exception, e:
            traceback.print_exc()
            return self.notFound

    def __nonzero__(self):
        return False


# Not found singleton
NotFound = FourOhFour(), ()
