# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from __future__ import generators
import warnings
import sys

# For backward compatability, provide lazy wrappers around the builtin renderers
def deprecationWrapper(name):
    def _(*args, **kwargs):
        warnings.warn(
            "stan.%s has been moved. Use rend.%s instead." % (name, name),
            DeprecationWarning, 2)
        from nevow import rend
        wrapped = getattr(rend, name)
        return wrapped(*args, **kwargs)
    return _

sequence = deprecationWrapper('sequence')
mapping = deprecationWrapper('mapping')
mergeSlot = deprecationWrapper('mergeSlot')

class Proto(str):
    """Proto is a string subclass. Instances of Proto, which are constructed
    with a string, will construct Tag instances in response to __call__
    and __getitem__, delegating responsibility to the tag.
    """
    __slots__ = []

    def __call__(self, **kw):
        return Tag(self)(**kw)

    def __getitem__(self, children):
        return Tag(self)[children]

    def fillSlots(self, slotName, slotValue):
        return Tag(self).fillSlots(slotName, slotValue)

    def clone(self, deep=True):
        return self

class xml(str):
    """Raw xml marker
    """
    __slots__ = []


class directive(object):
    """Marker for a directive in a template
    """
    __slots__ = ['name']
    def __init__(self, name):
        self.name = name
        
    def __repr__(self):
        return "directive('%s')" % self.name

class slot(object):
    """Marker for slot insertion in a template
    """
    __slots__ = ['name', 'children']
    def __init__(self, name):
        self.name = name
        self.children = []

    def __repr__(self):
        return "slot('%s')" % self.name

    def __getitem__(self, children):
        """Allow slots to have children. These children will not show up in the
        output, but they will be searched for patterns.
        """
        print "adding children to self", self, children
        if not isinstance(children, (list, tuple)):
            children = [children]
        self.children.extend(children)
        return self


class Tag(object):
    """Tag instances represent html tags with a tag name, attributes,
    and children. Tag instances can be constructed using the Prototype
    tags in the 'tags' module, or may be constructed directly with a tag
    name. Tags have two special methods, __call__ and __getitem__,
    which make representing trees of XML natural using pure python
    syntax. See the docstrings for these methods for more details.
    """
    specials = ['data', 'render', 'remember', 'pattern', 'key']

    slotData = None
    def __init__(self, tag, attributes=None, children=None, specials=None):
        self.tagName = tag
        if attributes is None:
            self.attributes = {}
        else:
            self.attributes = attributes
        if children is None:
            self.children = []
        else:
            self.children = children
        if specials is None:
            self._specials = {}
        else:
            self._specials = specials

    def fillSlots(self, slotName, slotValue):
        if self.slotData is None:
            self.slotData = {}
        self.slotData[slotName] = slotValue

    def __call__(self, **kw):
        """Change attributes of this tag. This is implemented using
        __call__ because it then allows the natural syntax:
        
        table(width="100%", height="50%", border="1")

        Attributes may be 'invisible' tag instances (so that
        a(href=invisible(data="foo", render=myhrefrenderer) works),
        strings, functions, or any other object which has a registered
        ISerializable adapter.
        
        A few magic attributes have values other than these, as they
        are not serialized for output but rather have special purposes
        of their own:
        
        data - The value is saved on the context stack and passed to
               render functions.
        render - A function to call that may modify the tag in any
                 way desired.
        remember - Remember the value on the context stack with
                   context.remember(value) for later lookup with
                   context.locate()
        pattern - Value should be a key that can later be used to locate
                  this tag with context.patternGenerator() or
                  context.allPatterns()
        key - A string used to give the node a unique label.
              This is automatically namespaced, so in:
              span(key="foo")[span(key="bar")]
              the inner span actually has a key of 'foo.bar'.
              The key is is intended for use as e.g. an html 'id' attribute,
              but will is not automatically output.
        """
        if not kw:
            return self

        for name in self.specials:
            if kw.has_key(name):
                setattr(self, name, kw[name])
                del kw[name]
            
        if kw.has_key('renderer'):
            if sys._getframe(1).f_code == Proto.__call__.func_code:
                stacklevel = 3
            else:
                stacklevel = 2
            warnings.warn("'renderer' attribute has been renamed 'render'",  DeprecationWarning, stacklevel)
            setattr(self, 'render', kw['renderer'])
            del kw['renderer']
        
        for k, v in kw.iteritems():
            if k[0] == '_':
                k = k[1:]
            self.attributes[k] = v
        return self

    def __getitem__(self, children):
        """Add children to this tag. Multiple children may be added by
        passing a tuple or a list. Children may be other tag instances,
        strings, functions, or any other object which has a registered
        ISerializable adapter.
        
        This is implemented using __getitem__ because it then allows
        the natural syntax:
        
        html[
            head[
                title["Hello World!"]
            ],
            body[
                "This is a page",
                h3["How are you!"],
                div(style="color: blue")["I hope you are fine."]
            ]
        ]
        """
        if not isinstance(children, (list, tuple)):
            children = [children]
        self.children.extend(children)
        return self

    def __iter__(self):
        """Prevent an infinite loop if someone tries to do
            for x in stantaginstance:
        """
        raise NotImplementedError, "Stan tag instances are not iterable."

    # FIXME: make this function actually be used.
    def precompilable(self):
        """Is this tag precompilable?
        
        Tags are precompilable if they will not be modified by a user
        render function.

        Currently, the following attributes prevent the tag from being
        precompiled:
        - render (because the function can modify its own tag)
        - pattern (because it is locatable and thus modifiable by an
                   enclosing renderer)
        """
        return self.render is Unset and self.pattern is Unset
    
    def _clone(self, obj, deep):
        if hasattr(obj, 'clone'):
            return obj.clone(deep)
        elif isinstance(obj, (list, tuple)):
            return [self._clone(x, deep)
                    for x in obj]
        else:
            return obj
        
    def clone(self, deep=True):
        """Return a clone of this tag. If deep is True, clone all of this
        tag's children. Otherwise, just shallow copy the children list
        without copying the children themselves.
        """
        if deep:
            newchildren = [self._clone(x, True) for x in self.children]
        else:
            newchildren = self.children[:]
        newattrs = self.attributes.copy()
        for key in newattrs:
            newattrs[key]=self._clone(newattrs[key], True)

        newslotdata = None
        if self.slotData:
            newslotdata = self.slotData.copy()
            for key in newslotdata:
                newslotdata[key] = self._clone(newslotdata[key], True)

        newtag = Tag(
            self.tagName,
            attributes=newattrs,
            children=newchildren,
            specials=self._specials.copy()
            )
        newtag.slotData = newslotdata
            
        return newtag

    def clear(self):
        """Clear any existing children from this tag.
        """
        self._specials = {}
        self.children = []
        return self

    def __repr__(self):
        rstr = ''
        if self.attributes:
            rstr += ', attributes=%r' % self.attributes
        if self._specials:
            rstr += ', specials=%r' % self._specials
        if self.children:
            rstr += ', children=%r' % self.children
        return "Tag(%r%s)" % (self.tagName, rstr)

    ## TODO: remove this deprecated api
    def _getRenderer(self):
        warnings.warn("'renderer' attribute has been renamed 'render'",  DeprecationWarning, 2)
        return self._specials.get('render', Unset)

    def _setRenderer(self):
        warnings.warn("'renderer' attribute has been renamed 'render'", DeprecationWarning, 2)
        self._specials['render'] = data

    renderer = property(_getRenderer, _setRenderer)

    def freeze(self):
        """Freeze this tag so that making future calls to __call__ or __getitem__ on the
        return value will result in clones of this tag.
        """
        def forever():
            while True:
                yield self.clone()
        from nevow.context import PatternTag
        return PatternTag(forever())


class UnsetClass:
    def __nonzero__(self):
        return False
    def __repr__(self):
        return "Unset"
Unset=UnsetClass()

def makeAccessors(special):
    def getSpecial(self):
        return self._specials.get(special, Unset)

    def setSpecial(self, data):
        self._specials[special] = data

    return getSpecial, setSpecial

for name in Tag.specials:
    setattr(Tag, name, property(*makeAccessors(name)))
del name


def specials(tag, special):
    """Generate tags with special attributes regardless of attribute value.
    """
    for childOrContext in getattr(tag, 'children', []):
        child = getattr(childOrContext, 'tag', childOrContext)
        
        if getattr(child, special, Unset) is not Unset:
            yield child
        else:
            for match in specials(child, special):
                yield match


def specialMatches(tag, special, pattern):
    """Generate special attribute matches starting with the given tag;
    if a tag has special, do not look any deeper below that tag, whether
    it matches pattern or not. Returns an iterable.
    """
    for childOrContext in getattr(tag, 'children', []):
        child = getattr(childOrContext, 'tag', childOrContext)
        
        data = getattr(child, special, Unset)
        if data == pattern:
            yield child
        elif data is Unset:
            for match in specialMatches(child, special, pattern):
                yield match

class CommentProto(Proto):
    __slots__ = []
    def __call__(self, **kw):
        return Comment(self)(**kw)

    def __getitem__(self, children):
        return Comment(self)[children]


class Comment(Tag):
    def __call__(self, **kw):
        return NotImplementedError('comments are not callable')

invisible = Proto('')


class Entity(object):
    def __init__(self, name, num, description):
        self.name = name
        self.num = num
        self.description = description

    def __repr__(self):
        return "Entity(%r, %r, %r)" % (self.name, self.num, self.description)

