
from twisted.web import static
from twisted.web.resource import Resource,IResource
from nevow import guard
from nevow import appserver

from twisted.cred.checkers import *
from twisted.cred.portal import *
from twisted.cred.credentials import *
from twisted.internet import address

from nevow.testutil import TestCase

class FakeHTTPChannel:
    # TODO: this should be an interface in twisted.protocols.http... lots of
    # things want to fake out HTTP
    def __init__(self):
        self.transport = self
        self.factory = self
        self.received_cookies = {}

    # 'factory' attribute needs this
    def log(self, req):
        pass

    # 'channel' of request needs this
    def requestDone(self, req):
        self.req = req

    # 'transport' attribute needs this
    def getPeer(self):
        return address.IPv4Address("TCP", "fake", 12345)
    def getHost(self):
        return address.IPv4Address("TCP", "fake", 80)

    def write(self, data):
        # print data
        pass
    def writeSequence(self, datas):
        for data in datas:
            self.write(data)

    # Utility for testing.

    def makeFakeRequest(self, path, username='',password=''):
        req = FakeHTTPRequest(self, queued=0)
        req.user = username
        req.password = password
        req.received_cookies.update(self.received_cookies)
        req.requestReceived("GET", path, "1.0")
        return req


class FakeHTTPRequest(appserver.NevowRequest):
    def __init__(self, *args, **kw):
        appserver.NevowRequest.__init__(self, *args, **kw)
        self._pchn = self.channel
        self._cookieCache = {}
        from cStringIO import StringIO
        self.content = StringIO()
        self.received_headers['host'] = 'fake.com'
        self.written = StringIO()

    def followRedirect(self):
        L = self.headers['location']
        if L.startswith('http://'):
            L = L[len("http://"):]
            urlist = L.split('/')
            urlist[0] = ''
            nexturi = '/'.join(urlist)
        else:
            nexturi = L
        return self._pchn.makeFakeRequest(nexturi)

    def followAllRedirects(self):
        R = self
        MAX_REDIRECTS = 5
        I = 1
        while R.headers.has_key('location'):
            assert I < MAX_REDIRECTS, "Too many redirects (to %s)" % R.headers['location']
            R = R.followRedirect()
            I += 1
        return R

    def write(self, data):
        self.written.write(data)
        appserver.NevowRequest.write(self, data)

    def addCookie(self, k, v, *args,**kw):
        appserver.NevowRequest.addCookie(self,k,v,*args,**kw)
        assert not self._cookieCache.has_key(k), "Should not be setting duplicate cookies!"
        self._cookieCache[k] = v
        self.channel.received_cookies[k] = v

    def processingFailed(self, fail):
        raise fail


class FakeSite(appserver.NevowSite):
    pass


class GuardTest(TestCase):

    def createPortal(self):
        r = SillyRealm()
        p = Portal(r)
        p.registerChecker(AllowAnonymousAccess(), IAnonymous)
        return p

    def createSessionWrapper(self, portal):
        swrap = guard.SessionWrapper(portal)
        self.sessions = swrap.sessions
        return swrap

    def createChannel(self, resource):
        s = FakeSite(resource)
        c = FakeHTTPChannel()
        c.site = s
        return c

    def testHttpAuthInit(self):
        p = self.createPortal()
        chan = self.createChannel(self.createSessionWrapper(p))
        p.registerChecker(InMemoryUsernamePasswordDatabaseDontUse(test='test'), IUsernamePassword)
        req = chan.makeFakeRequest("/", "test", "test")
        self.assertEquals(req.written.getvalue(), "Yes")
        req = chan.makeFakeRequest("/", "test", "test")
        req = chan.makeFakeRequest("/", "test", "test")
        self.assertEquals(len(self.sessions),1)

    def testSessionInit(self):
        p = self.createPortal()
        chan = self.createChannel(self.createSessionWrapper(p))

        # The first thing that happens when we attempt to browse with no session
        # is a cookie being set and a redirect being issued to the session url
        req = chan.makeFakeRequest('/xxx/yyy/')
        self.assertEquals( len(req._cookieCache.values()), 1, "Bad number of cookies in response.")
        # The redirect is set immediately and should have a path segment at the beginning matching our cookie
        self.failUnless(req.headers.has_key('location'))
        cookie = req._cookieCache.values()[0]

        # The URL should have the cookie segment in it and the correct path segments at the end
        self.assertEquals(req.headers['location'],
            'http://fake.com/%s/xxx/yyy/' % (guard.SESSION_KEY+cookie, ))

        # Now, let's follow the redirect
        req = req.followRedirect()
        # Our session should now be set up and we will be redirected to our final destination
        self.assertEquals(req.headers['location'].split('?')[0],
            'http://fake.com/xxx/yyy/')

        # Let's follow the redirect to the final page
        req = req.followRedirect()
        self.failIf(req.headers.has_key('location'))

        # We should have the final resource, which is an anonymous resource
        self.assertEquals(req.written.getvalue(), "No")

    sessions = {}

    def tearDown(self):
        for sz in self.sessions.values():
            sz.expire()

    def testUsernamePassword(self):
        p = self.createPortal()
        p.registerChecker(InMemoryUsernamePasswordDatabaseDontUse(test='test'), IUsernamePassword)
        chan = self.createChannel(self.createSessionWrapper(p))

        # Check the anonymous page
        req = chan.makeFakeRequest('/').followAllRedirects()
        self.assertEquals(req.written.getvalue(), "No")

        # Check the logged in page
        req = chan.makeFakeRequest('/__login__?username=test&password=test').followAllRedirects()
        self.assertEquals(req.written.getvalue(), "Yes")

        # Log out
        chan.makeFakeRequest("/__logout__").followRedirect()

        # Get the anonymous page again
        k = chan.makeFakeRequest("/")
        self.assertEquals(k.written.getvalue(), "No")

    def testLoginWithNoSession(self):
        p = self.createPortal()
        p.registerChecker(InMemoryUsernamePasswordDatabaseDontUse(test='test'), IUsernamePassword)
        chan = self.createChannel(self.createSessionWrapper(p))

        req = chan.makeFakeRequest('/__login__?username=test&password=test').followAllRedirects()
        self.assertEquals(req.written.getvalue(), "Yes")

    def testFormWithNoSession(self):
        p = self.createPortal()
        chan = self.createChannel(self.createSessionWrapper(p))

        req = chan.makeFakeRequest('/?aFormArgument=1').followAllRedirects()
        self.assertEquals(req.written.getvalue(), 'We got the form.')


class SillyAvatar(Resource):
    isLeaf = True
    def render(self, request):
        return 'Yes'


class SillyAnonymous(Resource):
    isLeaf = True
    def render(self, request):
        if request.args.get('aFormArgument'):
            return "We got the form."
        return 'No'


class SillyRealm:
    __implements__ = IRealm

    def requestAvatar(self, avatarId, mind, *interfaces):
        if avatarId == ANONYMOUS:
            return IResource, SillyAnonymous(), lambda: None
        else:
            return IResource, SillyAvatar(), lambda: None

